(*  Title:      GCM.thy
    Author:     Ludovic Henrio 
                2008

    Note:       
*)

header {* GCM Components *}


theory GCM imports FMap begin

subsection{* First defines Names *}
(* for simplicity Names are isomorphic to a finite subset of nats *)
consts max_name :: nat

typedef Name = "{n :: nat. n <= max_name }"
by auto;

lemma finite_Name_set: "finite {n :: nat. n <= max_name }";
apply (rule_tac n = "max_name + 1" and f = "% x. x" in nat_seg_image_imp_finite);
apply (simp add: ran_def);
apply force;
done;

lemma Univ_abs_name: "(UNIV :: (Name set)) = Abs_Name ` {n :: nat. n <= max_name }";
apply (subgoal_tac "! x :: Name. x : Abs_Name ` {n :: nat. n <= max_name }");
apply force;
apply (rule allI);
apply (subgoal_tac "Abs_Name(Rep_Name x) = x");
prefer 2;
apply (rule Rep_Name_inverse) ;
apply (erule subst);
apply (rule imageI);
apply (fold Name_def);
apply (rule Rep_Name);
done;

lemma finite_Name: "finite (UNIV :: (Name set))";
apply (subst Univ_abs_name);
apply (rule finite_imageI);
apply (rule finite_Name_set);
done;

instance Name :: "fintype" 
proof 
show "finite (UNIV :: (Name set))" by (rule finite_Name);
qed;

types       Fid = nat (*types for future ids *)

types Value = "nat\<times>Fid set" 
(* a nat symbolising the result + the list of  futures referenced by the value *)


datatype ItfKind= Client | Server

datatype Cardinality = Single

(* interface definition, types could be added here *)
record Interface = 
  kind::ItfKind
  card::Cardinality


datatype QualifiedName = This Name 
                         | Sub Name  Name (infix ".\<^sub>i" 50)

subsection{*Component and Requests structure *}
record Binding = 
  src:: QualifiedName 
  dest:: QualifiedName


record Request= 
id::Fid
parameter:: Value 
invokedItf:: Name

record Result= 
fid::Fid
fValue:: Value


datatype LCstate = Started | Stopped

(* behaviour of primitive components *)
datatype Action = Call Name Value Fid (*called interface + parameter + created id *)
                | ReceiveResult Fid Value (* future id x value received *)
  (*                | Wait Fid   future id --apparently not useful for the moment ... *)
                | EndService Fid Value (* value of the result for the request *)
                | NewService Name Value Fid (* name of the invoked interface, value of the parameter, corresponding future *)
                | Tau 

record intState = (* Primitive internal state *)
PNFstate:: LCstate
PRqRefs:: "Fid set" (* the list of request ids referenced by the state *)
Pstate:: nat (* a representation of the internal state: nat for simple *)
PcurrentReqs:: "Fid list" (*"Fid option" *)
         (* id of the current requests *)

constdefs reachable :: "(intState \<times> Action \<times> intState) set \<Rightarrow> intState set " 
          "reachable beh == {s'.\<exists> s a . (s,a,s')\<in>beh} "

typedef  Behaviours="{beh:: (intState \<times> Action \<times> intState) set. 
   (\<forall> s s' . ((s,Tau,s')\<in> beh \<longrightarrow> PRqRefs s'\<subseteq>PRqRefs s
                        \<and> (PcurrentReqs s') = (PcurrentReqs s))) \<and> (* tau operations can only forget futures *)
   (\<forall> s s' i v f. ((s,NewService i v f,s')\<in> beh \<longrightarrow> 
                        (\<forall> v' f'. \<exists> s' . (s,NewService i v' f',s')\<in>beh) \<and>
                        PRqRefs s'\<subseteq>PRqRefs s\<union>snd v \<and>
                        (PcurrentReqs s') = f#(PcurrentReqs s) ))\<and> 
    (\<forall> s s' i v f.((s,Call i v f,s')\<in>beh \<longrightarrow>( (* Call operations imply: *)
                        (\<forall> f'. \<exists> s'' . (s,Call i v f',s'')\<in>beh) (*any future id can be created*)
                       (*\<and>  i\<in>Clients does not work*)    (* invoked itf belong to the component *)   
                        \<and>(snd v)\<subseteq>PRqRefs s                 (* sent futures included in known futures *)
                        \<and>   PRqRefs s'\<subseteq> (PRqRefs s\<union>{f})
                        \<and> (PcurrentReqs s') = (PcurrentReqs s) ))) \<and>   (* only new future is f *)
   (\<forall> s\<in>reachable beh .\<forall> f. (\<forall> v. \<exists> s' . (s,ReceiveResult f v,s')\<in>beh)) \<and> 
                                                (* it is always possible to receive a future value for a  future *)
   (\<forall> s s' v f.((s,ReceiveResult f v,s')\<in>beh \<longrightarrow> (               (* ReceiveFuture implies: *)
                        PRqRefs s'\<subseteq> ((PRqRefs s)-{f})\<union>(snd v)
                        \<and> (PcurrentReqs s') = (PcurrentReqs s) 
                        \<and> (f\<notin>PRqRefs s \<longrightarrow>s=s') ))) \<and>
   (\<forall> s s' v f.((s,EndService f v,s')\<in>beh \<longrightarrow> (          (* EndService implies: *) 
                        (PRqRefs s'\<subseteq> PRqRefs s) (* EndService operations can only forget futures *)
                        \<and> (snd v \<subseteq> PRqRefs s)  (* sent value only contain known futures *)
                        \<and> f\<in>set (PcurrentReqs s)
                          \<and> (PcurrentReqs s') = remove1 f (PcurrentReqs s) )))        
                   }"
apply (unfold reachable_def);
by auto;

typedef WaitingStateFunction="{wait:: intState => Fid set. \<forall> s f. (f\<in> wait s  \<longrightarrow>f\<in>(PRqRefs s)) }"
by auto;

record PrimState = 
Pqueue:: "Request list" (* pending*)
PintState:: intState
PcomputedResults:: "Result list" (* could be a set *)
behaviour:: Behaviours (* the local semantics/behaviour of the component *)
                                                    (* the only infinite set *)
WaitingState:: WaitingStateFunction

definition Behaviour :: "PrimState \<Rightarrow> (intState\<times>Action\<times>intState) set"
where "Behaviour s \<equiv> Rep_Behaviours(behaviour s)"

record CompState = 
Cqueue:: "Request list" (* pending*)
CNFstate:: LCstate
CcomputedResults:: "Result list"

datatype Component =
             (* interfaces *)    (*sub components *) (* bindings: finite set *) (* state *)
  Primitive  "Name -~> Interface"                                     PrimState |
  Composite  "Name -~> Interface" "Name -~> Component" "Binding set" CompState



subsection{*Various basic functions*}
 constdefs Src_belongs :: "QualifiedName \<Rightarrow> Binding set \<Rightarrow> bool" (infix "\<in>\<^sub>b" 60) 
    "Src_belongs n b == n\<in> (\<lambda>x. src x) ` b"

(* to be changed if multicast interfaces ! *)
  constdefs NoDuplicateSrc:: "Binding set \<Rightarrow> bool" 
      "NoDuplicateSrc bindings==
        \<forall> s d1 d2.   ( \<lparr>src=s, dest=d1\<rparr>\<in> bindings\<and>\<lparr>src=s, dest=d2\<rparr>\<in> bindings) \<longrightarrow> d1=d2" 

primrec GetQualified:: "QualifiedName \<Rightarrow> Component \<Rightarrow> Interface option" where
  "GetQualified qn (Primitive itfs s) = None" |
  "GetQualified qn (Composite itfs subCp bindings s) = 
  (case qn of 
      (This n) \<Rightarrow>  (itfs n) |
      (CN.\<^sub>iIN) \<Rightarrow> (case (subCp CN) of
         None \<Rightarrow> None |
         Some Cp \<Rightarrow> (case Cp of
                        (Primitive itfs' s) \<Rightarrow> (itfs' IN) |
                        (Composite itfs' subCp' bindings' s) \<Rightarrow> (itfs' IN))))"

(* cplist returns the list of components taking part in a composition  order not deterministic !*)
primrec cpList:: "Component \<Rightarrow> Component list" and
        cpListoption:: "Component option \<Rightarrow> Component list" where
  "cpList (Primitive itfs s) = [(Primitive itfs s)]" |
  "cpList (Composite itfs subCp bindings s) =  (fold (op @)(\<lambda>x.(cpListoption (subCp x))) [Composite itfs subCp bindings s]  (dom subCp)) " |
  "cpListoption None =[]" |
  "cpListoption (Some c)=cpList c"

primrec ListFromOption:: " 'a option \<Rightarrow> 'a list" where
  "ListFromOption None = []" |
  "ListFromOption (Some a) = [a]"

lemma lengthListFromOption: "length (ListFromOption a)\<le>1"
by (case_tac a,auto);

(* rqidlist returens the list of request Ids in all components odf a composition order not deterministic !*)
primrec RqIdList:: "Component \<Rightarrow> Fid list" and
        RqIdListoption:: "Component option \<Rightarrow> Fid list" where
  "RqIdList (Primitive itfs s) = (map (\<lambda>r. fid r) (PcomputedResults s))
                                @ (PcurrentReqs (PintState s))
                                @(map (\<lambda>r. id r) (Pqueue s))"                |
  "RqIdList (Composite itfs subCp bindings s) =  (fold (op @)(\<lambda>x.(RqIdListoption (subCp x))) 
                                                 ((map (\<lambda>r. fid r) (CcomputedResults s))
                                                  @(map (\<lambda>r. id r) (Cqueue s)))
                                                 (dom subCp)) " |
  "RqIdListoption None =[]" |
  "RqIdListoption (Some c)=RqIdList c"


(* ReferencedRqs returens the set of request Ids referenced in all components*)
primrec ReferencedRqs::  "Component \<Rightarrow> Fid set" and
        ReferencedRqsoption:: "Component option \<Rightarrow> Fid set" where
  "ReferencedRqs (Primitive itfs s) = (foldl (\<lambda>x y. x\<union> snd(fValue y))
                                      (foldl (\<lambda>x y. x\<union> snd(parameter y)) (PRqRefs  (PintState s)) (Pqueue s)) (PcomputedResults s))" |
  "ReferencedRqs (Composite itfs subCp bindings s) =  (fold (op \<union>)(\<lambda>x.(ReferencedRqsoption (subCp x))) 
                                                 (foldl (\<lambda>x y. x\<union> snd (fValue y))
                                                 (foldl (\<lambda>x y. x\<union> snd (parameter y)) {} (Cqueue s)) (CcomputedResults s))
                                                 (dom subCp)) " |
  "ReferencedRqsoption None ={}" |
  "ReferencedRqsoption (Some c)=ReferencedRqs c"

(* ComputedRqSet returens the list of request Ids in all components of a composition order not deterministic !*)
primrec ComputedRqs::  "Component \<Rightarrow> Fid set" and
        ComputedRqsoption:: "Component option \<Rightarrow> Fid set" where
  "ComputedRqs (Primitive itfs s) = (set (map fid (PcomputedResults s)))" |
  "ComputedRqs (Composite itfs subCp bindings s) =  (fold (op \<union>)(\<lambda>x.(ComputedRqsoption (subCp x))) 
                                                 (set (map fid (CcomputedResults s)))
                                                 (dom subCp)) " |
  "ComputedRqsoption None ={}" |
  "ComputedRqsoption (Some c)=ReferencedRqs c"


primrec getQueue:: " Component \<Rightarrow> Request list" where 
 "getQueue (Primitive itfs s)=(Pqueue s)" |
 "getQueue (Composite itfs subCp bindings s)=(Cqueue s)"

primrec getNFstate:: " Component \<Rightarrow> LCstate" where 
 "getNFstate (Primitive itfs s)=(PNFstate (PintState s))" |
 "getNFstate (Composite itfs subCp bindings s)=(CNFstate s)"

primrec getComputedResults:: " Component \<Rightarrow> Result list" where 
 "getComputedResults (Primitive itfs s)=(PcomputedResults s)" |
 "getComputedResults (Composite itfs subCp bindings s)=(CcomputedResults s)"

primrec findInResultList:: "Result list \<Rightarrow> Fid \<Rightarrow> Value option" where
 "findInResultList [] n = None" |
 "findInResultList (R#RL) n = (if (fid R = n) then Some (fValue R) else findInResultList RL n)"

(* IS IT THE BEST DEFINITION? *)

primrec findResult:: " Component \<Rightarrow> Fid \<Rightarrow> Value option"
  and   findResultoption:: " Component option \<Rightarrow> Fid \<Rightarrow> Value option" where 
 "findResult (Primitive itfs s) n= (findInResultList (PcomputedResults s) n)" |
 "findResult (Composite itfs subCp bindings s) n=
     (case (findInResultList (CcomputedResults s) n) of
	Some x\<Rightarrow>Some x |
	None \<Rightarrow> if ran(\<lambda>x.(findResultoption (subCp x) n))={} then None 
                   else  
                   Some (contents (ran (\<lambda>x.(findResultoption (subCp x) n)))) )" |
 "findResultoption None n= None" |
 "findResultoption (Some cp) n=findResult cp n"

(* add lemmas HERE for uniqueness of result*)

primrec enQueue:: "Component \<Rightarrow> Request \<Rightarrow> Component" (infix "\<leftarrow>" 70) where
 "enQueue (Primitive itfs s) R = (Primitive itfs (s\<lparr>Pqueue:=(Pqueue s) @ [R]\<rparr>))" |
 "enQueue (Composite itfs subCp bindings s) R = (Composite itfs subCp bindings (s\<lparr>Cqueue:=(Cqueue s) @ [R]\<rparr>))"


subsection{*define correct components *}
(* only considers single interfaces *)
(* each binding has a single source.
   each binding corresponds to existing source and dest
   and recursively for each component
  *) 
primrec CorrectComponentStructure :: "Component \<Rightarrow> bool" 
and CorrectComponentStructureoption:: "Component option \<Rightarrow> bool"  where
  "CorrectComponentStructure (Primitive itfs s) = True" |
  "CorrectComponentStructure (Composite itfs subCp bindings cs) = ((\<forall> b\<in> bindings. 
     (GetQualified (src b) (Composite itfs subCp bindings cs) =Some\<lparr>kind=Client,card=Single\<rparr>)
      \<and> (GetQualified (dest b) (Composite itfs subCp bindings cs)=Some\<lparr>kind=Server,card=Single\<rparr>))
     \<and> NoDuplicateSrc bindings
     \<and>(fold (op &) (\<lambda> x. CorrectComponentStructureoption(subCp x)) True (dom subCp)))" |
  " CorrectComponentStructureoption None = True" |
  " CorrectComponentStructureoption (Some cp) = CorrectComponentStructure cp" 

(* a cp is correct if each request id is unique 
   and each referenced request exists
   and the component structure is correct *)

(*constdefs CorrectBehaviour

constdefs CorrectBehaviour = " (intState \<times> Action \<times> intState) set \<Rightarrow> bool"
   "CorrectBehaviour (Primitive itfs s) = 
" |
  "CorrectBehaviour (Composite itfs subCp bindings cs) = ((\<forall> b\<in> bindings. 
*)

constdefs CorrectComponent:: "Component\<Rightarrow>bool" 
"CorrectComponent c == CorrectComponentStructure c \<and> distinct (RqIdList c) \<and> (ReferencedRqs c) \<subseteq>(set (RqIdList c))"

(* !!!!!!!! ADD destination interface of request exist + idem for Calls !!!!!!!!!! *)
(* !!!!!!!! ADD all interfaces are plugged NOT MANDATORY !!!!!!!!!! *)




subsection{*lemmas on basic functions*}

(* Axioms for the moment! to see *)

axioms CpList_rqIdList_Q: " \<exists> m. id ((getQueue (cpList c!i))!n) = (RqIdList c)!m"
axioms CpList_rqIdList_R: " \<exists> m. fid ((getComputedResults (cpList c!i))!n) = (RqIdList c)!m"
axioms CpList_rqIdList_C: " (cpList c!i) = (Primitive itfs s) \<and> ((PcurrentReq s) = (Some i))\<Longrightarrow> \<exists>m.  (RqIdList c)!m=i"


end

